@@ -28,6 +28,11 @@
 #define DMDEBUG(x...)
 #endif
 
+/*
+ * The percentage increment we will wake up users at
+ */
+#define WAKE_UP_PERCENT 5
+
 /*
  * Hard sector size used all over the kernel
  */
@@ -86,18 +91,6 @@ struct origin {
 	struct list_head snapshots;
 };
 
-/*
- * Useful macro for running the store functions.  Use
- * store_int_fn if you want the return value.
- */
-#define store_fn(snap, fn, args...) \
-    if ((snap)->store. ## fn) \
-        (snap)->store. ## fn ( &(snap)->store , ## args )
-
-#define store_int_fn(snap, fn, args...) \
-    (((snap)->store. ## fn) ? \
-        ((snap)->store. ## fn ( &(snap)->store , ## args )) : 0)
-
 /*
  * Size of the hash table for origin volumes. If we make this
  * the size of the minors list then it should be nearly perfect
@@ -215,7 +208,7 @@ static int init_exception_table(struct exception_table *et, uint32_t size)
 	int i;
 
 	et->hash_mask = size - 1;
-	et->table = vmalloc(sizeof(struct list_head) * (size));
+	et->table = vcalloc(size, sizeof(struct list_head));
 	if (!et->table)
 		return -ENOMEM;
 
@@ -225,7 +218,7 @@ static int init_exception_table(struct exception_table *et, uint32_t size)
 	return 0;
 }
 
-static void exit_exception_table(struct exception_table *et, kmem_cache_t * mem)
+static void exit_exception_table(struct exception_table *et, kmem_cache_t *mem)
 {
 	struct list_head *slot, *entry, *temp;
 	struct exception *ex;
@@ -285,7 +278,13 @@ static struct exception *lookup_exception(struct exception_table *et,
 
 static inline struct exception *alloc_exception(void)
 {
-	return kmem_cache_alloc(exception_cache, GFP_NOIO);
+	struct exception *e;
+
+	e = kmem_cache_alloc(exception_cache, GFP_NOIO);
+	if (!e)
+		e = kmem_cache_alloc(exception_cache, GFP_ATOMIC);
+
+	return e;
 }
 
 static inline void free_exception(struct exception *e)
@@ -382,7 +381,7 @@ static int init_hash_tables(struct dm_snapshot *s)
 
 /*
  * Construct a snapshot mapping: <origin_dev> <COW-dev> <p/n>
- * <chunk-size> <extent-size>
+ * <chunk-size>
  */
 static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 			int argc, char **argv, void **context)
@@ -492,12 +491,15 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 
 	s->chunk_size = chunk_size;
 	s->chunk_mask = chunk_size - 1;
+	s->type = *persistent;
 	for (s->chunk_shift = 0; chunk_size;
 	     s->chunk_shift++, chunk_size >>= 1)
 		;
 	s->chunk_shift--;
 
 	s->valid = 1;
+	s->last_percent = 0;
+	s->table = t;
 	init_rwsem(&s->lock);
 
 	/* Allocate hash table for COW data */
@@ -511,12 +513,11 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 	 * Check the persistent flag - done here because we need the iobuf
 	 * to check the LV header
 	 */
-#if 0
+	s->store.snap = s;
+
 	if ((*persistent & 0x5f) == 'P')
-		r = dm_create_persistent(&s->store, s, blocksize,
-					 extent_size, context);
+		r = dm_create_persistent(&s->store, s->chunk_size);
 	else
-#endif
 		r = dm_create_transient(&s->store, s, blocksize, context);
 
 	if (r) {
@@ -525,13 +526,6 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 		goto bad_free1;
 	}
 
-	/* Allocate the COW iobuf and set associated variables */
-	r = store_int_fn(s, init, blocksize, extent_size, context);
-	if (r) {
-		*context = "Couldn't initialise exception store";
-		goto bad_free1;
-	}
-
 	/* Flush IO to the origin device */
 #if LVM_VFS_ENHANCEMENT
 	fsync_dev_lockfs(s->origin->dev);
@@ -548,12 +542,13 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 #if LVM_VFS_ENHANCEMENT
 	unlockfs(s->origin->dev);
 #endif
+	kcopyd_inc_client_count();
 
 	*context = s;
 	return 0;
 
       bad_free2:
-	store_fn(s, destroy);
+	s->store.destroy(&s->store);
 
       bad_free1:
 	exit_exception_table(&s->pending, pending_cache);
@@ -574,17 +569,21 @@ static void snapshot_dtr(struct dm_table *t, void *context)
 {
 	struct dm_snapshot *s = (struct dm_snapshot *) context;
 
+	dm_table_event(s->table);
+
 	unregister_snapshot(s);
 
 	exit_exception_table(&s->pending, pending_cache);
 	exit_exception_table(&s->complete, exception_cache);
 
 	/* Deallocate memory used */
-	store_fn(s, destroy);
+	s->store.destroy(&s->store);
 
 	dm_table_put_device(t, s->origin);
 	dm_table_put_device(t, s->cow);
 	kfree(s);
+
+	kcopyd_dec_client_count();
 }
 
 /*
@@ -630,22 +629,19 @@ static void error_buffers(struct buffer_head *bh)
 	}
 }
 
-/*
- * Called when the copy I/O has finished
- */
-static void copy_callback(int err, void *context)
+static void pending_complete(struct pending_exception *pe, int success)
 {
-	struct pending_exception *pe = (struct pending_exception *) context;
-	struct dm_snapshot *s = pe->snap;
 	struct exception *e;
+	struct dm_snapshot *s = pe->snap;
 
-	if (!err) {
-		/* Update the metadata if we are persistent */
-		store_fn(s, commit_exception, &pe->e);
-
+	if (success) {
 		e = alloc_exception();
 		if (!e) {
-			/* FIXME: what do we do now ? */
+			printk("Unable to allocate exception.");
+			down_write(&s->lock);
+			s->store.drop_snapshot(&s->store);
+			s->valid = 0;
+			up_write(&s->lock);
 			return;
 		}
 
@@ -653,30 +649,41 @@ static void copy_callback(int err, void *context)
 		 * Add a proper exception, and remove the
 		 * inflight exception from the list.
 		 */
-		down_write(&pe->snap->lock);
+		down_write(&s->lock);
 
 		memcpy(e, &pe->e, sizeof(*e));
 		insert_exception(&s->complete, e);
 		remove_exception(&pe->e);
 
 		/* Submit any pending write BHs */
-		up_write(&pe->snap->lock);
+		up_write(&s->lock);
 
 		flush_buffers(pe->snapshot_bhs);
 		DMDEBUG("Exception completed successfully.");
 
+		/* Notify any interested parties */
+		if (s->store.percent_full) {
+			int pc = s->store.percent_full(&s->store);
+
+			if (pc >= s->last_percent + WAKE_UP_PERCENT) {
+				dm_table_event(s->table);
+				s->last_percent = pc - pc % WAKE_UP_PERCENT;
+			}
+		}
+
 	} else {
 		/* Read/write error - snapshot is unusable */
 		DMERR("Error reading/writing snapshot");
 
-		down_write(&pe->snap->lock);
-		store_fn(pe->snap, drop_snapshot);
-		pe->snap->valid = 0;
+		down_write(&s->lock);
+		s->store.drop_snapshot(&s->store);
+		s->valid = 0;
 		remove_exception(&pe->e);
-		up_write(&pe->snap->lock);
+		up_write(&s->lock);
 
 		error_buffers(pe->snapshot_bhs);
 
+		dm_table_event(s->table);
 		DMDEBUG("Exception failed.");
 	}
 
@@ -688,6 +695,30 @@ static void copy_callback(int err, void *context)
 	free_pending_exception(pe);
 }
 
+static void commit_callback(void *context, int success)
+{
+	struct pending_exception *pe = (struct pending_exception *) context;
+	pending_complete(pe, success);
+}
+
+/*
+ * Called when the copy I/O has finished.  kcopyd actually runs
+ * this code so don't block.
+ */
+static void copy_callback(int err, void *context)
+{
+	struct pending_exception *pe = (struct pending_exception *) context;
+	struct dm_snapshot *s = pe->snap;
+
+	if (err)
+		pending_complete(pe, 0);
+
+	else
+		/* Update the metadata if we are persistent */
+		s->store.commit_exception(&s->store, &pe->e, commit_callback,
+					  pe);
+}
+
 /*
  * Dispatches the copy operation to kcopyd.
  */
@@ -745,7 +776,7 @@ static struct pending_exception *find_pending_exception(struct dm_snapshot *s,
 		pe->snap = s;
 		pe->started = 0;
 
-		if (store_int_fn(s, prepare_exception, &pe->e)) {
+		if (s->store.prepare_exception(&s->store, &pe->e)) {
 			free_pending_exception(pe);
 			s->valid = 0;
 			return NULL;
@@ -797,7 +828,7 @@ static int snapshot_map(struct buffer_head *bh, int rw, void *context)
 			pe = find_pending_exception(s, bh);
 
 			if (!pe) {
-				store_fn(s, drop_snapshot);
+				s->store.drop_snapshot(&s->store);
 				s->valid = 0;
 			}
 
@@ -832,6 +863,20 @@ static int snapshot_map(struct buffer_head *bh, int rw, void *context)
 	return r;
 }
 
+static void list_merge(struct list_head *l1, struct list_head *l2)
+{
+	struct list_head *l1_n, *l2_p;
+
+	l1_n = l1->next;
+	l2_p = l2->prev;
+
+	l1->next = l2;
+	l2->prev = l1;
+
+	l2_p->next = l1_n;
+	l1_n->prev = l2_p;
+}
+
 static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 {
 	int r = 1;
@@ -866,13 +911,13 @@ static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 		if (!e) {
 			pe = find_pending_exception(snap, bh);
 			if (!pe) {
-				store_fn(snap, drop_snapshot);
+				snap->store.drop_snapshot(&snap->store);
 				snap->valid = 0;
-			} else {
 
+			} else {
 				if (last)
-					list_splice(&pe->siblings,
-						    &last->siblings);
+					list_merge(&pe->siblings,
+						   &last->siblings);
 
 				last = pe;
 				r = 0;
@@ -901,6 +946,43 @@ static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 	return r;
 }
 
+static int snapshot_status(status_type_t type, char *result,
+			   int maxlen, void *context)
+{
+	struct dm_snapshot *snap = (struct dm_snapshot *) context;
+	char cow[16];
+	char org[16];
+
+	switch (type) {
+	case STATUSTYPE_INFO:
+		if (!snap->valid)
+			snprintf(result, maxlen, "Invalid");
+		else {
+			if (snap->store.percent_full)
+				snprintf(result, maxlen, "%d%%",
+					 snap->store.percent_full(&snap->
+								  store));
+			else
+				snprintf(result, maxlen, "Unknown");
+		}
+		break;
+
+	case STATUSTYPE_TABLE:
+		/*
+		 * kdevname returns a static pointer so we need
+		 * to make private copies if the output is to
+		 * make sense.
+		 */
+		strncpy(cow, kdevname(snap->cow->dev), sizeof(cow));
+		strncpy(org, kdevname(snap->origin->dev), sizeof(org));
+		snprintf(result, maxlen, "%s %s %c %ld", org, cow,
+			 snap->type, snap->chunk_size);
+		break;
+	}
+
+	return 0;
+}
+
 /*
  * Called on a write from the origin driver.
  */
@@ -966,12 +1048,31 @@ static int origin_map(struct buffer_head *bh, int rw, void *context)
 	return (rw == WRITE) ? do_origin(dev, bh) : 1;
 }
 
+static int origin_status(status_type_t type, char *result,
+			 int maxlen, void *context)
+{
+	struct dm_dev *dev = (struct dm_dev *) context;
+
+	switch (type) {
+	case STATUSTYPE_INFO:
+		result[0] = '\0';
+		break;
+
+	case STATUSTYPE_TABLE:
+		snprintf(result, maxlen, "%s", kdevname(dev->dev));
+		break;
+	}
+
+	return 0;
+}
+
 static struct target_type origin_target = {
 	name:	"snapshot-origin",
 	module:	THIS_MODULE,
 	ctr:	origin_ctr,
 	dtr:	origin_dtr,
 	map:	origin_map,
+	status:	origin_status,
 	err:	NULL
 };
 
@@ -981,6 +1082,7 @@ static struct target_type snapshot_target = {
 	ctr:	snapshot_ctr,
 	dtr:	snapshot_dtr,
 	map:	snapshot_map,
+	status:	snapshot_status,
 	err:	NULL
 };
 
